/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */ 

#include <ANN/ANN.h>// ANN declarations

#include "BufferedNode.h" 
#include <iostream>
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include <math.h>
#include "Map.h"

using namespace std;
using namespace HARK;

class EgoNoiseEstimation;

DECLARE_NODE(EgoNoiseEstimation)
/*Node
 *  
 * @name EgoNoiseEstimation
 * @category HARK:ENE
 * @description This node estimates ego noise using templates, creates and updates a database.
 *
 * @input_name SPECTRUM
 * @input_type Map<int,ObjectRef>
 * @input_description Complex spectrum for each channel. Key:ChannelID, Value:Data for each channel.
 *
 * @input_name SEP_SPECTRUM
 * @input_type Map<int,ObjectRef>
 * @input_description Separated complex spectrum for each source. Key:SrcID, Value:Data for each source.
 *
 * @input_name POSITION
 * @input_type Vector<float>
 * @input_description position information (joint angle).
 *
 * @input_name VELOCITY
 * @input_type Vector<float>
 * @input_description velocity information (joint angular velocity).
 *
 * @output_name ESTIMATED_NOISE
 * @output_type Matrix<complex<float> >
 * @output_description Estimate noise.
 *
 * @parameter_name FFT_LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description FFT length in sample.
 *
 * @parameter_name NCHANNELS
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description The number of input channels.
 *
 * @parameter_name TEMPLATE_DB_FILENAME
 * @parameter_type string
 * @parameter_description Filename of the existing template, no filename means no template database.
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Print debug messages.
 *
 * @parameter_name SEARCH_WINDOW_WIDTH
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description The number of best nearest neighbors, which are subject to a cross correlation (CC) in terms of their spectra. Most similar template in terms of its features and spectrum is selected. 1 means, CC is not applied.
 *
 * @parameter_name K
 * @parameter_type int
 * @parameter_value 10
 * @parameter_description Number of nearest neighbors (K-NN).  
 *
 * @parameter_name EPS
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_description Error bound for nearest neighbors.  
 *
 * @parameter_name MAX_POINTS
 * @parameter_type int
 * @parameter_value 250000
 * @parameter_description Maximum number of data points for nearest neighbors. Reduce this number to save the memory.
 *
 * @parameter_name ENABLE_NORMALIZATION
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Normalize the joint-related information (features). F_new=(F_old-F_min)/(F_max-F_min). 
 *
 * @parameter_name POSITION_MAX
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_NORMALIZATION=true
 * @parameter_description Maximum value of each element in POSITION input. This should have the same number of elements as that of POSITION input.
 *
 * @parameter_name POSITION_MIN
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_NORMALIZATION=true
 * @parameter_description Minimum value of each element in POSITION input. This should have the same number of elements as that of POSITION input.
 *
 * @parameter_name VELOCITY_MAX
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_NORMALIZATION=true
 * @parameter_description Maximum value of each element in VELOCITY input. This should have the same number of elements as that of VELOCITY input.
 *
 * @parameter_name VELOCITY_MIN
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_NORMALIZATION=true
 * @parameter_description Minimum value of each element in VELOCITY input. This should have the same number of elements as that of VELOCITY input.
 *
 * @parameter_name ENABLE_VELOCITY_OVERWRITE
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Overwrite the velocities provided by the robot with the calculated values from positions.
 *
 * @parameter_name FILTER_SIZE
 * @parameter_type int
 * @parameter_value 3
 * @parameter_valid ENABLE_VELOCITY_OVERWRITE=true
 * @parameter_description Filter size towards future (symmetric to the past).
 *
 * @parameter_name POSITION_WEIGHT
 * @parameter_type float
 * @parameter_value 1
 * @parameter_description Weight of the position for NN distance calculation.
 *
 * @parameter_name VELOCITY_WEIGHT
 * @parameter_type float
 * @parameter_value 1
 * @parameter_description Weight of the velocity for NN distance calculation.
 *
 * @parameter_name ENABLE_LEARNING
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Update the internal template database using incremental learning.
 *
 * @parameter_name TEMPLATE_DB_NEW_FILENAME
 * @parameter_type string
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_description filename of the new database.
 *
 * @parameter_name LEARNING_INTERVAL
 * @parameter_type int
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_value 5
 * @parameter_description Time interval [in frames] to re-calculate the nearest neighborhood tree (must be greater than K). It is a trade-off between learning speed and computational power.
 *
 * @parameter_name TIME_CONSTANT
 * @parameter_type float
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_value 10.0
 * @parameter_description Time constant for smoothing (in frames).
 *
 * @parameter_name ERROR_MARGIN
 * @parameter_type float
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_value 0.0001
 * @parameter_description Error threshold for nearest neighbor, which triggers the decision of a new template or update of an old template.
 *
 * @parameter_name ENABLE_NOISE_DETECTION
 * @parameter_type bool
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Detects external directional noises and stops learning.
 *
 * @parameter_name ENABLE_ADAPTIVE
 * @parameter_type bool
 * @parameter_valid ENABLE_LEARNING=true
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Adaptive forgetting factor which is time and frequency variant. True: Adaptive, False: Fixed
 *
 * @parameter_name TILT_FEATURES
 * @parameter_type float
 * @parameter_valid ENABLE_ADAPTIVE=true
 * @parameter_value 1000
 * @parameter_description Tilt value for the update formula in terms of feature similarity.
 *
 * @parameter_name TILT_SPECTRA
 * @parameter_type float
 * @parameter_valid ENABLE_ADAPTIVE=true
 * @parameter_value 10
 * @parameter_description Tilt value for the update formula in terms of spectra similarity.
 *
 END*/

class EgoNoiseEstimation : public BufferedNode 
{

  ////////////////////////////////////////////////////////////////////
  //////////        DECLARATIONS OF VARIABLES         ////////////////
  ////////////////////////////////////////////////////////////////////
  int m_iOutputID, m_iOutput2ID, m_iOutput3ID, m_iOutput4ID, m_iOutput5ID, m_iOutput6ID, m_iSpectrumID, m_iSepSpectrumID;
  int m_iPositionID, m_iVelocityID;
  int nChannels; //number of channels
  int numTrainingData;//number of training templates 
  float epsilon; //error margin
  float forgettingFactor, ff, ff1, ff2; //forgetting factor related variables
  float timeConstant; // time constant for forgetting factor
  float wPos; //position weight
  float wVel; //velocity weight
  float weight, totalWeight; //temp weight value holder 
  int searchWindowWidth; //search window width
  int learningInterval; //learning interval
  int databaseCounter; // internal counter for database elements
  Matrix<float> matSpectrumR; //Matrix of template spectrum (real part)
  Matrix<float> matSpectrumI; //Matrix of template spectrum (imaginary part)
  Matrix<float> matNormalization; //matrix containing the normalization values
  int fftlength; //fft length
  int filterSize; // used for joint status smoothing
  int nFreqBins; // the half of the real fft value for practical computing purposes
  int totalFreqBins; //nChannels*NumberOfFrequencyBins
  int totalTemplateWidth; //totalTemplateWidth = nFeatures+nChannels*NumberOfFrequencyBins
  vector<float> currFeatures; //internal variable to store the joint status
  float tilt, tilt2;
  string newRootFilename;
  bool enable_normalize, enable_debug, enable_learning, enable_adaptive, enable_overwrite, enable_NoiseDetection;
  Vector<float> position_max;
  Vector<float> position_min;
  Vector<float> velocity_max;
  Vector<float> velocity_min;  

  //ANN parameters
  int k;// number of nearest neighbors
  int nFeatures;// dimension
  int nPositions;// dimension of position
  double eps; // error bound
  int maxPts; // maximum number of data points
  ANNpointArray dataPts;// data points
  ANNpoint queryPt;// query point
  ANNidxArray nnIdx;// near neighbor indices
  ANNdistArray dists;// near neighbor distances
  ANNkd_tree* kdTree;// search structure
    
  typedef Matrix<complex<float> > FCMatrix;
  typedef Matrix<float> FMatrix;
  
public:
  EgoNoiseEstimation(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    ////////////////////////////////////////////////////////////////////
    //////////        INITIALIZATION OF VARIABLES         //////////////
    ////////////////////////////////////////////////////////////////////
    m_iSpectrumID = addInput("SPECTRUM");
    m_iSepSpectrumID = addInput("SEP_SPECTRUM");
    m_iPositionID = addInput("POSITION");
    m_iVelocityID = addInput("VELOCITY");
    m_iOutput2ID = addOutput("ESTIMATED_NOISE");
    m_iOutputID = addOutput("PROCESSED_SPECTRUM"); //disabled:Output spectrum
    m_iOutput3ID = addOutput("INSTANTANEOUS_SNR");// disabled:current SNR averaged over all frequency bins.
    m_iOutput4ID = addOutput("ESTIMATED_JOINTSTATUS"); //disabled:Estimated joint status 
    m_iOutput5ID = addOutput("PROCESSED_JOINTSTATUS"); //disabled:Input joint status after being processed (smoothed) internally
    m_iOutput6ID = addOutput("CONFIDENCE_VALUES"); //disabled:Confidence values||forgetting factors
    nChannels = dereference_cast<int>(parameters.get("NCHANNELS"));
    timeConstant = dereference_cast<float>(parameters.get("TIME_CONSTANT"));
    searchWindowWidth = dereference_cast<int>(parameters.get("SEARCH_WINDOW_WIDTH"));
    wPos = dereference_cast<float>(parameters.get("POSITION_WEIGHT"));
    wVel = dereference_cast<float>(parameters.get("VELOCITY_WEIGHT"));
    learningInterval = dereference_cast<int>(parameters.get("LEARNING_INTERVAL"));
    filterSize = dereference_cast<int>(parameters.get("FILTER_SIZE"));
    inputsCache[m_iPositionID].lookAhead = filterSize;
    inputsCache[m_iPositionID].lookBack = filterSize;
    inputsCache[m_iVelocityID].lookAhead = filterSize;
    inputsCache[m_iVelocityID].lookBack = filterSize;
    k = dereference_cast<int>(parameters.get("K"));
    epsilon = dereference_cast<float>(parameters.get("ERROR_MARGIN"));
    fftlength = dereference_cast<int>(parameters.get("FFT_LENGTH"));
    nFreqBins = fftlength / 2 + 1;
    totalFreqBins = nChannels*nFreqBins;
    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
    enable_learning = dereference_cast<bool>(parameters.get("ENABLE_LEARNING"));
    enable_adaptive = dereference_cast<bool>(parameters.get("ENABLE_ADAPTIVE"));
    enable_overwrite = dereference_cast<bool>(parameters.get("ENABLE_VELOCITY_OVERWRITE"));
    tilt = dereference_cast<float>(parameters.get("TILT_FEATURES"));
    tilt2 = dereference_cast<float>(parameters.get("TILT_SPECTRA"));
    enable_normalize = dereference_cast<bool>(parameters.get("ENABLE_NORMALIZATION")) ;
    position_max = object_cast<Vector<float> >(parameters.get("POSITION_MAX"));
    position_min = object_cast<Vector<float> >(parameters.get("POSITION_MIN"));
    velocity_max = object_cast<Vector<float> >(parameters.get("VELOCITY_MAX"));
    velocity_min = object_cast<Vector<float> >(parameters.get("VELOCITY_MIN"));
    enable_NoiseDetection = dereference_cast<bool>(parameters.get("ENABLE_NOISE_DETECTION"));
    maxPts = dereference_cast<int>(parameters.get("MAX_POINTS")); // maximum number of data points
    eps = (double)dereference_cast<float>(parameters.get("EPS")); // error bound
    forgettingFactor = exp( -1/timeConstant); // forgetting factor dt = 1 frame;  tc=0.2 -> ff = 0.067; tc = 1 -> ff= 0.36; tc=3 -> ff = 0.75; default: tc=10 -> ff= 0.9; tc = 100 -> 0.99
    databaseCounter = 0; //the instantaneous length of the database   
    inOrder = true;        
  }
    
  ~EgoNoiseEstimation()//destruct the module
  {
    if (enable_learning && parameters.exist("TEMPLATE_DB_NEW_FILENAME")) // Save whatever remains in the data structures to the external datafile
      {      
        newRootFilename = object_cast<string>(parameters.get("TEMPLATE_DB_NEW_FILENAME"));
	FILE *fp;
	string newFilename = newRootFilename + "R";
	if((fp=fopen(newFilename.c_str(), "wb"))==NULL) 
	  {
	    printf("Cannot open file.\n");
	    exit(1);
	  }
	int m_iDataSize = 0;
	fwrite(&m_iDataSize, sizeof(int), 1, fp);
        int iDataType = 0;
	fwrite(&iDataType, sizeof(int), 1, fp);
	int numrow = databaseCounter-1+numTrainingData;
	fwrite(&numrow, sizeof(int), 1, fp);
	int numcol = nFeatures + totalFreqBins;
	fwrite(&numcol, sizeof(int), 1, fp);
	for (int ind = 0 ;  ind < databaseCounter-1+numTrainingData ; ind++)
	  {
	    for (int j = 0 ; j < nFeatures ; j++)
	      {
		if (j < nPositions) weight = wPos; else weight = wVel;
		float fval = dataPts[ind][j]/weight;
		fwrite(&fval, sizeof(float), 1, fp);
	      }
	    for (int i = 0 ; i < totalFreqBins ; i++)
	      {
		float fval = matSpectrumR(ind,i);
		fwrite(&fval, sizeof(float), 1, fp);
	      }
	  }
	fclose(fp);
  		  
	newFilename = newRootFilename + "I";
	if((fp=fopen(newFilename.c_str(), "wb"))==NULL) 
	  {
	    printf("Cannot open file.\n");
	    exit(1);
	  }
	m_iDataSize = 0;
	fwrite(&m_iDataSize, sizeof(int), 1, fp);
        iDataType = 0;
	fwrite(&iDataType, sizeof(int), 1, fp);
	numrow = databaseCounter-1+numTrainingData;
	fwrite(&numrow, sizeof(int), 1, fp);
	numcol = nFeatures + totalFreqBins;
	fwrite(&numcol, sizeof(int), 1, fp);
	for (int ind = 0 ;  ind < databaseCounter-1+numTrainingData ; ind++)
	  {
	    for (int j = 0 ; j < nFeatures ; j++)
	      {
		if (j < nPositions) weight = wPos; else weight = wVel;
		float fval = dataPts[ind][j]/weight;
		fwrite(&fval, sizeof(float), 1, fp);
	      }
	    for (int i = 0 ; i < totalFreqBins ; i++)
	      {
		float fval = matSpectrumI(ind,i);
		fwrite(&fval, sizeof(float), 1, fp);
	      }
	  }
	fclose(fp);
      }
  }
    
  virtual void initialize()
  {
    outputs[m_iOutputID].lookAhead = outputs[m_iOutput2ID].lookAhead = outputs[m_iOutput4ID].lookAhead = outputs[m_iOutput5ID].lookAhead = outputs[m_iOutput6ID].lookAhead = max(max(max(max(outputs[m_iOutputID].lookAhead, outputs[m_iOutput2ID].lookAhead), outputs[m_iOutput4ID].lookAhead), outputs[m_iOutput5ID].lookAhead), outputs[m_iOutput6ID].lookAhead);
    outputs[m_iOutputID].lookBack = outputs[m_iOutput2ID].lookBack = outputs[m_iOutput4ID].lookBack = outputs[m_iOutput5ID].lookBack = outputs[m_iOutput6ID].lookBack = max(max(max(max(outputs[m_iOutputID].lookBack, outputs[m_iOutput2ID].lookBack), outputs[m_iOutput4ID].lookBack), outputs[m_iOutput5ID].lookBack), outputs[m_iOutput6ID].lookBack);
    this->BufferedNode::initialize();
  }
    
  virtual int translateInput(string inputName)
  {
    for (unsigned int i = 0; i < inputs.size(); i++) {
      if (inputs[i].name == inputName) {
	return i;
      }
    }
  }
  
  void calculate(int output_id, int count, Buffer &out)
  {
    int i, j, finalPos, nSources;
    vector<float> absTemp, argTemp, realTemp, imagTemp, G, logPower;
    float tmpCorr, tmpCorrDenom, finalCorr;
    complex<float> c1(0,-1);
    ObjectRef orInputSpectrum, orInputSpectrum2, orInputTemp, orOutputTemp,  orOutputTemp6, orConfidenceValues, orTemplateTemp; //Object references for all inputs and outputs
    ObjectRef orPositionTemp, orVelocityTemp;
    Map<int, ObjectRef>::iterator pmTemp; //iterator for all channels and then speakers
      

    ///////////////////////////////////////////////////////////////////////////
    ////////////////////       DEFINING INPUTS AND OUTPUTS        /////////////
    ///////////////////////////////////////////////////////////////////////////
    // Input 1 Spectrum
    orInputSpectrum = getInput(m_iSpectrumID, count);
    Map<int, ObjectRef> &mpInputValue = object_cast<Map<int, ObjectRef> >(orInputSpectrum);
    // Input 2 Spectrum
    orInputSpectrum2 = getInput(m_iSepSpectrumID, count);
    Map<int, ObjectRef> &mpsepSpectrumValue = object_cast<Map<int, ObjectRef> >(orInputSpectrum2);
    // Input 3 Position
    orPositionTemp = getInput(m_iPositionID, count);
    Vector<float> &mPositionValue = object_cast<Vector<float> >(orPositionTemp);
    // Input 4 Velocity      
    orVelocityTemp = getInput(m_iVelocityID, count);
    Vector<float> &mVelocityValue = object_cast<Vector<float> >(orVelocityTemp);

    if(count == 0){
      nFeatures = mPositionValue.size()+mVelocityValue.size();
      nPositions = mPositionValue.size();

      if(enable_normalize){
	matNormalization.resize(2, nFeatures);// Normalization matrix for the joint status
	if (position_max.size() + velocity_max.size() != nFeatures) {
	  throw new NodeException(this, string("Size mismatched : POSITION_MAX + VELOCITY_MAX != POSITION + VELOCITY."), __FILE__, __LINE__);
	}
	if (position_min.size() + velocity_min.size() != nFeatures) {
	  throw new NodeException(this, string("Size mismatched : POSITION_MIN + VELOCITY_MIN != POSITION + VELOCITY."), __FILE__, __LINE__);
	}
	for (i = 0; i < mPositionValue.size(); i++){
	  matNormalization(0,i) = position_min[i]; matNormalization(1,i) = position_max[i]; 	  
	}
	for (i = 0; i < mVelocityValue.size(); i++){
	  matNormalization(0,i+mPositionValue.size()) = velocity_min[i]; matNormalization(1,i) = velocity_max[i]; 	  
	}
      }

      // Declarations for ANN
      FILE *m_pfin;	// file token for reading and writing 
      float *pfDataTemp, *pfDataPos; //internal variables to store the content of the read templates
      int iRtn;
      currFeatures.resize(nFeatures);
      queryPt = annAllocPt(nFeatures);// allocate query point
      dataPts = annAllocPts(maxPts, nFeatures);// allocate data points
      nnIdx = new ANNidx[k];// allocate near neigh indices
      dists = new ANNdist[k];// allocate near neighbor dists
                      
      ////////////////////////////////////////////////////////////////////
      //////////        READ THE R TEMPLATE DATABASE FROM FILE       /////
      ////////////////////////////////////////////////////////////////////          
      if (parameters.exist("TEMPLATE_DB_FILENAME"))// 1) If an old template database exists, read it
	{
	  string rootFilename = object_cast<string>(parameters.get("TEMPLATE_DB_FILENAME"));        
	  string filename = rootFilename + "R";
	  m_pfin = fopen(filename.c_str(),"rb");
	  if (!m_pfin)
	    {
	      throw(new NodeException(NULL, "Template file 'R' cannot be found.", __FILE__, __LINE__));
	    }
	  int m_iDataSize = 0;
	  fread(&m_iDataSize, sizeof(int), 1, m_pfin);
	  int iDataType;
	  fread(&iDataType, sizeof(int), 1, m_pfin);
	  iRtn = fread(&numTrainingData, sizeof(int), 1, m_pfin);
	  iRtn = fread(&totalTemplateWidth, sizeof(int), 1, m_pfin);
	  if (totalTemplateWidth!=(totalFreqBins+nFeatures))
	    {
	      cout <<  "Number of training data: "<<numTrainingData << "\nNumber of template width: " <<totalTemplateWidth<< " must be equal to " << totalFreqBins+nFeatures<<"\n" ;    
	      throw(new NodeException(NULL, "Sizes do not match", __FILE__, __LINE__));
	    }
	  matSpectrumR.resize(maxPts, totalFreqBins); // Matrix containing the real spectrum data
	  pfDataTemp = 0;
	  pfDataTemp = new float[totalTemplateWidth];
	  for (i = 0 ; i < numTrainingData ; i++)
	    {          
	      iRtn = fread(pfDataTemp, sizeof(float), totalTemplateWidth, m_pfin);
	      if (iRtn != totalTemplateWidth)
		{
		  if (pfDataTemp)
		    {
		      delete [] pfDataTemp;
		      pfDataTemp = 0;
		    }
		  throw(new NodeException(NULL, "Template database size is invalid.", __FILE__, __LINE__));
		  return;
		} 
	      pfDataPos = pfDataTemp;
	      for (j = 0 ; j < totalTemplateWidth ; j++)
		{
		  if (j<nFeatures) //Assign the features, e.g. first 64
		    {
		      if (j < nPositions) weight = wPos; else weight = wVel;
		      dataPts[i][j] = weight* *pfDataPos;
		    }
		  else //Assign the spectra, e.g. last 257
		    {
		      matSpectrumR(i, j-nFeatures) = *pfDataPos; //read the spectrum matrix
		    }
		  pfDataPos++; // increment to the next data in the template
		  
		}
	    }           	              
	  if (pfDataTemp)
	    {
	      delete [] pfDataTemp;
	      pfDataTemp = 0;
	    }
	  fclose(m_pfin);
	  ////////////////////////////////////////////////////////////////////
	  //////////        READ THE I TEMPLATE DATABASE FROM FILE       /////
	  ////////////////////////////////////////////////////////////////////        
	  filename = rootFilename + "I";
	  m_pfin = fopen(filename.c_str(),"rb");
	  if (!m_pfin)
	    {
	      throw(new NodeException(NULL, "Template file 'I' cannot be found.", __FILE__, __LINE__));
	    }
	  fread(&m_iDataSize, sizeof(int), 1, m_pfin);
	  fread(&iDataType, sizeof(int), 1, m_pfin);
	  iRtn = fread(&numTrainingData, sizeof(int), 1, m_pfin);
	  iRtn = fread(&totalTemplateWidth, sizeof(int), 1, m_pfin);      
	  if (totalTemplateWidth!=(totalFreqBins+nFeatures))
	    {
	      cout <<  "Number of training data: "<<numTrainingData << "\nNumber of template width: " <<totalTemplateWidth<< " must be equal to " << totalFreqBins+nFeatures<<"\n" ;    
	      throw(new NodeException(NULL, "Sizes do not match", __FILE__, __LINE__));
	    }     
	  matSpectrumI.resize(maxPts, totalFreqBins); // Matrix containing the real spectrum data
	  pfDataTemp = 0;
	  pfDataTemp = new float[totalTemplateWidth];
	  for (i = 0 ; i < numTrainingData ; i++)
	    {          
	      iRtn = fread(pfDataTemp, sizeof(float), totalTemplateWidth, m_pfin);
	      if (iRtn != totalTemplateWidth)
		{
		  if (pfDataTemp)
		    {
		      delete [] pfDataTemp;
		      pfDataTemp = 0;
		    }
		  throw(new NodeException(NULL, "Template database size is invalid.", __FILE__, __LINE__));
		  return;
		} 
	      pfDataPos = pfDataTemp;
	      for (j = 0 ; j < totalTemplateWidth ; j++)
		{
		  if (j<nFeatures) //Assign the features, e.g. first 64
		    {
		      if (j < nPositions) weight = wPos; else weight = wVel;
		      dataPts[i][j] = weight* *pfDataPos;
		    }
		  else //Assign the spectra, e.g. last 257
		    {
		      matSpectrumI(i, j-nFeatures) = *pfDataPos; //read the spectrum matrix
		    }
		  pfDataPos++; // increment to the next data in the template
		  
		}
	    }           	              
	  if (pfDataTemp)
	    {
	      delete [] pfDataTemp;
	      pfDataTemp = 0;
	    }
	  fclose(m_pfin);             
	  ////////////////////////////////////////////////////////////////////
	  //////////        Create the initial NN tree from the database    //
	  ////////////////////////////////////////////////////////////////////
	  kdTree = new ANNkd_tree(// build search structure
				  dataPts,	// the data points
				  numTrainingData-1,		// number of points (-1 because of nPts++ operation)
				  nFeatures);		// dimension of space
	  
	}
      else// 2) If no database exists, start to create a new one
	{
	  printf("WARNING: Unable to load a template database.\nCreating a new template database...\n");
	  numTrainingData = 0;        
	  matSpectrumR.resize(maxPts, totalFreqBins); 
	  matSpectrumI.resize(maxPts, totalFreqBins);
	  
	  for (i = 0 ; i < k ; i++)//initialize the vectors
	    {
	      nnIdx[i]=0;
	      dists[i]=1; //some arbitrary value greater than epsilon
	      for (j = 0; j < nFeatures; j++)
		{
		  dataPts[i][j] = 10; //some arbitrary value greater than epsilon
		}
	    }
	  //Create the initial tree from the database
	  kdTree = new ANNkd_tree(// build search structure
				  dataPts,	// the data points
				  k,
				  nFeatures);		// dimension of space
	}
      
#if 0
      cout << "SizeR:" << matSpectrumR.nrows() << ", " << matSpectrumR.ncols() << endl;
      // for(int i1 = 0; i1 < matSpectrumR.nrows(); i1++){
      //   for(int i2 = 0; i2 < matSpectrumR.ncols(); i2++){
      // 	cout << matSpectrumR(i1,i2) << " ";
      //   }
      //   cout << endl;
      // }    
      for(int i2 = 0; i2 < matSpectrumR.ncols(); i2++){
	cout << matSpectrumR(0,i2) << " ";
      }
      cout << endl;
      cout << "SizeI:" << matSpectrumI.nrows() << ", " << matSpectrumI.ncols() << endl;
      // for(int i1 = 0; i1 < matSpectrumI.nrows(); i1++){
      //   for(int i2 = 0; i2 < matSpectrumI.ncols(); i2++){
      // 	cout << matSpectrumI(i1,i2) << " ";
      //   }
      //   cout << endl;
      // }    
      for(int i2 = 0; i2 < matSpectrumI.ncols(); i2++){
	cout << matSpectrumI(0,i2) << " ";
      }
      cout << endl;
      cout << "SizeP:" << maxPts << ", " << nFeatures << endl;
      // for(int i1 = 0; i1 < maxPts; i1++){
      //   for(int i2 = 0; i2 < nFeatures; i2++){
      // 	cout << dataPts[i1][i2] << " ";
      //   }
      //   cout << endl;
      // }    
      for(int i2 = 0; i2 < nFeatures; i2++){
	cout << dataPts[0][i2] << " ";
      }
      cout << endl;
#endif

    }

    // Output1 Processed Spectrum
    Buffer &outputBuffer = *(outputs[m_iOutputID].buffer) ;
    orOutputTemp = ObjectRef(new Map<int, ObjectRef>);
    Map<int, ObjectRef> &mpOutputValue = object_cast<Map<int, ObjectRef> >(orOutputTemp);
    outputBuffer[count] = orOutputTemp;
    // Output2 Predicted Noise
    RCPtr<FCMatrix> estimatedTemplate(new FCMatrix(nChannels, nFreqBins));    
    (*(outputs[m_iOutput2ID].buffer))[count] = estimatedTemplate; 
    // Output4 Predicted Joint Status
    RCPtr<FMatrix> orJointStatusOutput(new FMatrix(1, nFeatures));    
    (*(outputs[m_iOutput4ID].buffer))[count] = orJointStatusOutput;
    // Output5 Processed Joint Status
    RCPtr<FMatrix> orpJointStatusOutput(new FMatrix(1, nFeatures));    
    (*(outputs[m_iOutput5ID].buffer))[count] = orpJointStatusOutput; 
    // Output6 Confidence values
    Buffer &confidenceValuesBuffer = *(outputs[m_iOutput6ID].buffer) ;
    orConfidenceValues = ObjectRef(new Map<int,ObjectRef>);
    Map<int,ObjectRef> &mConfidenceValues = object_cast<Map<int,ObjectRef> >(orConfidenceValues);
    confidenceValuesBuffer[count] = orConfidenceValues;
     

    ///////////////////////////////////////////////////////////////////////////
    //////////////        JOINT STATUS MANIPULATIONS       ////////////////////
    ///////////////////////////////////////////////////////////////////////////
    if (enable_overwrite) // Overwrite the velocities and copy to a temporary vector
      {
        for(j = 0; j < mPositionValue.size(); j++){
	  currFeatures[j] = mPositionValue[j];
	}
        for(j = 0; j < mVelocityValue.size(); j++){
	  currFeatures[j+mPositionValue.size()] = 0;
	}
	
        for (i = -filterSize; i <= filterSize; i++) 
	  {            
	    orVelocityTemp = getInput(m_iVelocityID, max(0,count+i));
	    Vector<float> &mVelocityValue2 = object_cast<Vector<float> >(orVelocityTemp);
	    for(j = 0; j < mVelocityValue2.size(); j++)
	      {
		currFeatures[j+mPositionValue.size()] = currFeatures[j+mPositionValue.size()] + (mVelocityValue2[j]/(float)(2*filterSize+1)); //Apply moving average for the velocities
	      }
	  }
      }
    else //Copy the jointstatus input to a temporary vector
      {      
        for(j = 0; j < mPositionValue.size(); j++){
	  currFeatures[j] = mPositionValue[j];
	}
        for(j = 0; j < mVelocityValue.size(); j++){
	  currFeatures[j+mPositionValue.size()] = mVelocityValue[j];
	}
      }
    // Normalize the inputs and create a queryPt
    if (enable_normalize)
      {
      	for (j = 0 ; j < nFeatures ; j++)
	  {
	    if (j < nPositions) weight = wPos; else weight = wVel;
	    queryPt[j] = weight * min(1.0f,max(0.0f,(currFeatures[j]-matNormalization(0,j))/(matNormalization(1,j)-matNormalization(0,j)))); //With normalization
	  }
      }
    else
      {
      	for (j = 0 ; j < nFeatures ; j++)
	  {
	    if (j < nPositions) weight = wPos; else weight = wVel;
	    queryPt[j] = weight * currFeatures[j]; //Without normalization
	  }
      }      
    ///////////////////////////////////////////////////////////////////////////
    //////////////          TEMPLATE ESTIMATION            ////////////////////
    ///////////////////////////////////////////////////////////////////////////      
      
    // Calculate the confidence of the instantaneous queryPt, closest indices and dists       
    kdTree->annkSearch(// search
		       queryPt,	// query point
		       k,		// number of near neighbors
		       nnIdx,	// nearest neighbors (returned)
		       dists,	// distance (returned)
		       eps);		// error bound
    if (enable_debug)
      {
        cout << "***********************************************************";
        cout << "\n\ncount: "<< count;
        cout << "\nQuery Point: ";
        for (j = 0 ; j < nFeatures ; j++) cout << queryPt[j]<<" ";
	cout<<"\nEstimated Point: ";
        for (j = 0 ; j < nFeatures ; j++) cout << dataPts[nnIdx[0]][j]<<" ";
	cout<<"\n\n"; 
	for (int tmpcntr = 0; tmpcntr < k; tmpcntr++)	cout << "Index: "<< nnIdx[tmpcntr] << "\tDistance: " << dists[tmpcntr] << "\n";	
      }
      
    ///////////////////////////////////////////////////////////////////////////
    //////////////          TEMPLATE LEARNING LOOP         ////////////////////
    ///////////////////////////////////////////////////////////////////////////
      
    nSources = mpsepSpectrumValue.size() ; // number of sources
    int currCh = 0;
    for (pmTemp = mpInputValue.begin() ; pmTemp != mpInputValue.end() ; pmTemp++) //each channel is
      {
        // assign input data pointers
        orInputTemp = pmTemp->second;
        Vector<complex<float> > &vcfInputTemp = object_cast<Vector<complex<float> > >(orInputTemp);
        
        //instantiate the temporary data holders   
        absTemp.resize(nFreqBins); //abs of the current spectrum
        argTemp.resize(nFreqBins); //arg of the current spectrum
        realTemp.resize(nFreqBins); //real of the current spectrum
        imagTemp.resize(nFreqBins); //imag of the current spectrum
        
        ///////////////////////////////////////////////////////////////////////////
        //////////////          SPECTRA CORRELATION            ////////////////////
        ///////////////////////////////////////////////////////////////////////////        
    
        //Perform a correlation operation between the current noisy signal and adjacent templates of the estimated template
        // if searchWindowWidth == 1 -> this loop is skipped automatically
        finalCorr = 10000;//An arbitrarily chosen maximum value
        finalPos = 0;
        for (int windowCntr = 0; windowCntr < searchWindowWidth; windowCntr++)
	  {
	    tmpCorr = 0; tmpCorrDenom = 0;
	    for (i = 0 ; i < nFreqBins ; i++)
	      {
		absTemp[i] = abs(vcfInputTemp[i]);
		argTemp[i] = arg(vcfInputTemp[i]);
		realTemp[i] = real(vcfInputTemp[i]);
		imagTemp[i] = imag(vcfInputTemp[i]);
		tmpCorr = tmpCorr + abs(absTemp[i] - abs(complex<float> (matSpectrumR(nnIdx[windowCntr],currCh*nFreqBins+i), matSpectrumI(nnIdx[windowCntr],currCh*nFreqBins+i))));
		tmpCorrDenom = tmpCorrDenom + absTemp[i];
	      }
	    tmpCorr = tmpCorr/(tmpCorrDenom*nFreqBins);
	    if (tmpCorr <= finalCorr)
	      {
          	finalCorr = tmpCorr;
          	finalPos = windowCntr; //finalPos yields the index of the closest template taking the spectral correlation into account 
	      }
	  }
        
        if (enable_debug && currCh == 0)
	  {
	    cout << "\nQuery Spectrum (0-th channel): ";
	    for (j = 0 ; j < nFreqBins ; j++) cout <<absTemp[j]<<" ";
     	    cout<<"\nEstimated Spectrum (0-th channel): ";
	    for (j = 0 ; j < nFreqBins ; j++) cout << abs(complex<float> (matSpectrumR(nnIdx[finalPos],currCh*nFreqBins+j), matSpectrumI(nnIdx[finalPos],currCh*nFreqBins+j)))<<" ";
    	    cout<<"\n\n"; 
    	  }

        ///////////////////////////////////////////////////////////////////////////
        //////////////            IDWA CALCULATION             ////////////////////
        ///////////////////////////////////////////////////////////////////////////        
                
        // Calculation for the normalization factor of IDWA
        totalWeight = 0;        
        if (k == 1) // if k == 1, apply no IDWA
	  {
	    finalPos = 0;
	  } 
        else
	  {
	    for (int tmpCntr = 0;tmpCntr<k;tmpCntr++)
	      {
		totalWeight = totalWeight + (1.0/max(dists[tmpCntr],0.00000001)); //totalWeight = the normalization constant
	      }
	  }
        ///////////////////////////////////////////////////////////////////////////
        //////////////        ONLINE LEARNING OF TEMPLATES     ////////////////////
        ///////////////////////////////////////////////////////////////////////////
        
        if(enable_learning)
	  {   
	    ///////////////////////////////////////////////////////////////////////////
	    //////////////        UPDATE OF THE TEMPLATES          ////////////////////
	    ///////////////////////////////////////////////////////////////////////////

	    if ( (enable_NoiseDetection && !nSources) || !enable_NoiseDetection)// Either  the threshold to detect the interfering noise sources is higher or the noise detection is inactive
	      {
		//Online learning loop
		if ((count == 0) || (dists[0]>epsilon) || epsilon==0) //1) insert template to the database
		  {
		    for (j = 0 ; j < nFeatures ; j++) dataPts[databaseCounter+numTrainingData][j] = queryPt[j];
		    // include the latest spectral vector into the database of matSpectrum
		    for (i = 0 ; i < nFreqBins ; i++) 
		      {
			matSpectrumR(databaseCounter+numTrainingData,currCh*nFreqBins+i)=realTemp[i];
			matSpectrumI(databaseCounter+numTrainingData,currCh*nFreqBins+i)=imagTemp[i];
		      }
		    if (currCh == nChannels-1) databaseCounter++; //this assures that the counter is increased only once for each pmTemp
		  }
		else //2) Update Database
		  {
		    if (enable_adaptive && enable_NoiseDetection) //it means (dists[0]<epsilon)
		      {
			ff1 = 1/ (1.0f + exp(-tilt * (dists[0])));// forgetting factor related to error of the features
			ff2 = 1/ (1.0f + exp(-tilt2 * finalCorr));// forgetting factor related to error of the spectra
			ff = (ff1+ff2)/2; // averaging -> ff=[0.5 1]
			if (enable_debug)                
			  {
			    cout << count << " ff=" << ff<< ", ff1="<< ff1 << ", finalCorr="<< " " << finalCorr << ", ff2="<< " " << ff2 <<endl;
			  }       			    
		      }
		    else // 
		      {
			ff = forgettingFactor;
		      }
		    for (i = 0 ; i < nFreqBins ; i++) 
		      {
			matSpectrumR(nnIdx[finalPos],currCh*nFreqBins+i)= ff * matSpectrumR(nnIdx[finalPos],currCh*nFreqBins+i) + (1-ff) *realTemp[i]; //update the real part of the template
			matSpectrumI(nnIdx[finalPos],currCh*nFreqBins+i)= ff * matSpectrumI(nnIdx[finalPos],currCh*nFreqBins+i) + (1-ff) *imagTemp[i]; //update the imaginary part of the template
		      }
		  }
	      }
	    // Send the output mConfidenceValues
	    orOutputTemp6 = ObjectRef(Vector<float>::alloc(3));        
	    Vector<float> &vfConfidenceValues = object_cast<Vector<float> >(orOutputTemp6);
	    vfConfidenceValues[0]= ff1; vfConfidenceValues[1]= ff2; vfConfidenceValues[2]= ff;
	    mConfidenceValues.insert(make_pair(pmTemp->first, orOutputTemp6));
	  }    
      	currCh++; // end of ptemp	  		
      }

    // Send the outputs of estimatedTemplate, orJointStatusOutput and orpJointStatusOutput
    for (i = 0 ; i < nFreqBins ; i++)
      {        
        for (j = 0 ; j < nChannels ; j++)
	  { 
	    (*estimatedTemplate)(j, i) = complex<float> (matSpectrumR(nnIdx[0],j*nFreqBins+i), matSpectrumI(nnIdx[0],j*nFreqBins+i));
	  }
      }  
    for (i = 0 ; i < nFeatures ; i++)
      {
        (*orJointStatusOutput)(0,i)= dataPts[nnIdx[0]][i];// Send the estimated joint status to the output
        (*orpJointStatusOutput)(0,i)= queryPt[i];
      }  
      
    ///////////////////////////////////////////////////////////////////////////
    //////////////        RE_BUILDING THE DATABASE         ////////////////////
    ///////////////////////////////////////////////////////////////////////////
    if(enable_learning) // rebuild the database once in a while because it is costly
      {             
        if ( (enable_NoiseDetection && !nSources) || (!enable_NoiseDetection))// 
	  {
	    // re-build the tree with the dataPts
	    if (count % learningInterval == 0)
	      {
		kdTree = new ANNkd_tree(// build search structure
					dataPts,	// the data points
					max(k,numTrainingData-1+databaseCounter),		// number of points (-1 because of nPts++ operation)(+databaseCounter because of the valuable entries only)
					nFeatures);		// dimension of space
		cout << "New tree built with "<< numTrainingData+databaseCounter<< " templates from " <<numTrainingData+count <<" actual templates\n";
	      }
	  }
      }
  }
  IN_ORDER_NODE_SPEEDUP(EgoNoiseEstimation);
};

